local class = require("silo.lib.class")

local Peripherals = class("silo.Peripherals")

function Peripherals:new(ctx)
  self._activeFilter = false

  local queue = {}
  self.queue = queue

  local curTask, curFilter = nil
  local mainThread = coroutine.create(function()
    while true do
      if curTask then
        local filter = curTask.peripheral
        if filter == true then filter = "all peripherals"
        elseif filter == false then filter = "no peripheral"
        elseif type(filter) == "table" then
          filter = {}
          for k in pairs(curTask.peripheral) do filter[#filter + 1] = k end
          filter = table.concat(filter, " ")
        end

        curTask:fn()

        curTask, curFilter = nil, nil
      else
        os.pullEvent()
      end
    end
  end)

  ctx:addTask(function()
    while true do
      local event = table.pack(coroutine.yield())
      -- Do we need to republish this event?

      if curTask == nil then
        curTask = table.remove(queue, 1)
      end

      if curTask and (curFilter == nil or event[1] == curFilter or event[1] == "terminate") then
        self._activeFilter = curTask.peripheral
        local ok, res = coroutine.resume(mainThread, table.unpack(event, 1, event.n))
        self._activeFilter = false

        if not ok then error(res, 0) end

        curFilter = res
      end
    end
  end)
end

function Peripherals:isEnabled(name)
  local filter = self._activeFilter
  if filter == false then return false end
  if filter == true or filter == name then return true end
  if type(filter) == "table" and filter[name] then return true end
  return false
end

function Peripherals:wrap(name)
  local wrapped = peripheral.wrap(name)
  if not wrapped then
    error("Cannot wrap peripheral '" .. name .. "'")
  end

  local out = { _name = name }
  for method, func in pairs(wrapped) do
    out[method] = function(...)
      if not self:isEnabled(name) then
        error("Peripheral '" .. name .. "' is not enabled", 2)
      end

      local res = table.pack(pcall(func, ...))
      if res[1] then
        return table.unpack(res, 2, res.n)
      else
        error(("%s for (%s.%s)"):format(res[2], name, method))
      end
    end
  end

  return out
end

function Peripherals:execute(task)
  if type(task) ~= "table" then error("bad argument #1 (table expected") end
  if type(task.fn) ~= "function" then error("bad key 'fn' (function expected)") end

  if task.priority == nil then task.priority = 0 end
  if task.unique ~= true then task.unique = false end
  if task.peripheral == nil then task.peripheral = false end

  local queue = self.queue
  local inserted = false

  if task.unique then
    for i = 1, #queue do
      if queue[i].fn == task.fn then
        if queue[i].priority < task.priority then
          table.remove(queue, i)
          break
        else
          return
        end
      end
    end
  end

  if not inserted then
    for i = 1, #queue do
      if queue[i].priority < task.priority then
        table.insert(queue, i, task)
        inserted = true
        break
      end
    end
  end

  if not inserted then
    queue[#queue + 1] = task
  end

  if #queue == 1 then os.queueEvent("task_enqueue") end
end

return Peripherals
